package cz.mendelu.dnaAnalyser.batch;

import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.SequenceRepository;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataRepository;
import cz.mendelu.dnaAnalyser.user.User;
import cz.mendelu.dnaAnalyser.utils.exception.NotFoundException;
import cz.mendelu.dnaAnalyser.utils.exception.NoContentException;
import cz.mendelu.dnaAnalyser.utils.model.Identifiable;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import javax.naming.NotContextException;
import java.text.MessageFormat;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.function.Consumer;
import java.util.function.UnaryOperator;

import static org.springframework.data.domain.Sort.unsorted;


/**
 * Created by Jan Kolomazník on 30.6.17.
 */
@Slf4j
@Service
public class BatchService {

    @Value("${batch.asynchronous}")
    private boolean asynchronous;

    @Value("${batch.threadPoolSize}")
    private int threadPoolSize;

    @Autowired
    private BatchRepository batchRepository;

    @Autowired
    private SequenceRepository sequenceRepository;

    private ExecutorService executor;

    @PostConstruct
    public void initExecutorService() {
        if (asynchronous) {
            log.info("Start new Executor service for batch analysis");
            executor = (threadPoolSize > 0)
                    ? Executors.newWorkStealingPool(threadPoolSize)
                    : Executors.newWorkStealingPool();
        }
    }

    public Optional<Batch> findBatchFor(Identifiable entity) {
        Class entityClass = entity.getClass();
        UUID entityId = entity.getId();
        try {
            Optional<Batch> optionalBatch = batchRepository.findBatchByEntityClassAndEntityId(entityClass.getName(), entityId);
            return optionalBatch;
        } catch (Exception e) {
            log.warn(String.format("Batch object for %s(id: %s) not found. Reason: %s", entityClass, entityId, e.getMessage()));
            return Optional.empty();
        }
    }

    public BatchBuilder createBatch(final User owner, final Identifiable identifiable) {
        final String entityClass = identifiable.getClass().getName();
        final UUID entityId = identifiable.getId();
        final String name = MessageFormat.format("{0}[{1}]", entityClass, entityId);
        return new BatchBuilder() {

            private Consumer<UUID> operator;

            @Override
            public BatchBuilder execute(Consumer<UUID> operator) {
                this.operator = operator;
                return this;
            }

            @Override
            public Batch start() {
                if(entityClass == null){
                    throw new IllegalArgumentException("Batch entityClass can not be null.");
                }
                if(entityId == null){
                    throw new IllegalArgumentException("Batch entityId can not be null.");
                }
                if(owner == null){
                    throw new IllegalArgumentException("Batch owner can not be null.");
                }
                log.info("Create batch {}.", name);
                Batch batch = new Batch();
                batch.setEntityClass(entityClass);
                batch.setEntityId(entityId);
                batch.setOwner(owner);
                batch.setName(name);
                batch.setStatus(Batch.Status.CREATED);
                batch.setCreated(new Date());
                batchRepository.save(batch);
                if (asynchronous) {
                    runBatchInPool(batch, operator);
                } else {
                    runBatch(batch, operator);
                }
                return batch;
            }
        };
    }

    private void runBatchInPool(Batch batch, Consumer<UUID> operator) {
        batch.setStatus(Batch.Status.WAITING);
        batchRepository.save(batch);

        // Test if executor service is up and ready
        if (executor == null || executor.isShutdown() || executor.isTerminated()) {
            log.warn("Executor service is donw, create new instance!");
            initExecutorService();
        }
        executor.execute(() -> runBatch(batch, operator));
    }

    private void runBatch(Batch batch, Consumer<UUID> operator) {
            log.info("Start batch {} for {}[{}].", batch.getName(), batch.getEntityClass(), batch.getEntityId());
            batch.setStatus(Batch.Status.RUNNING);
            batch.setStarted(new Date());
            batchRepository.save(batch);
            try {
                // FIXME User data id been hold in memory, not load from db!
                SecurityContextHolder.getContext().setAuthentication(new BatchAuthentication(batch));
                long cpuTimeBeforeTaskRun = ThreadCPUTime.cpuTime();
                operator.accept(batch.getEntityId());
                long cpuTimeAfterTaskRun = ThreadCPUTime.cpuTime();
                batch.setCpuTime(cpuTimeAfterTaskRun - cpuTimeBeforeTaskRun);
                log.info("Finish batch {}.", batch.getName());
                batch.setStatus(Batch.Status.FINISH);
                batch.setFinished(new Date());
                batchRepository.save(batch);
            } catch (Throwable t) {
                log.error(String.format("Batch %s: '%s' failed.", batch.getEntityId(), batch.getName()), t);
                batch.setStatus(Batch.Status.FAILED);
                batch.setException(t.getMessage());
                batch.setFinished(new Date());
                batchRepository.save(batch);
            }

    }

    public Iterable<Batch> findAllBatches() {
        return batchRepository.findAll();
    }

    public Iterable<Batch> findAllByEntityClassAndOwner(String entityClass) {
        return batchRepository.findAllByEntityClass(entityClass);
    }

    public Batch findBatchByEntityClassAndEntityId(String entityClass, UUID entityId){

        Optional<Batch> batch = batchRepository.findBatchByEntityClassAndEntityId(entityClass, entityId);
        Optional<Sequence> sequence = sequenceRepository.findById(entityId);
        if(!batch.isPresent() && !sequence.isPresent())
            throw new NotFoundException("Batch for entity '%s' with id: %s", entityClass, entityId);
        if(!batch.isPresent() && sequence.isPresent()){
            throw new NoContentException("Batch for entity '%s' with id: %s", entityClass, entityId);
        }
        return batch.get();
    }

    void deleteBatchByEntityClassAndEntityId(String entityClass, UUID entityId) {
        batchRepository.deleteBatchByEntityClassAndEntityId(entityClass, entityId);
    }
}
